import os
import os.path
import getpass
from pyrustic import manager
from pyrustic.manager.handler.build_handler import BuildHandler
from pyrustic.manager.core.funcs import create_kurl
from jayson import Jayson


class PublishHandler:
    """
    Description
    -----------
    Use this command to publish the latest distribution
    package previously built with the command 'build'.

    Usage
    -----
    - Description: Publish
    - Command: publish

    Hooking
    -------
    This command will run hooks if they exist.
    The legal hooks are:
    - pre_publishing_hook.py
    - post_publishing_hook.py
    """

    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._pre_publishing_hook = None
        self._post_publishing_hook = None
        self._process(target, app_pkg)

    def _process(self, target, app_pkg):
        if target is None:
            print("Please link a Target first. Check 'help target'.")
            return
        if not os.path.exists(os.path.join(target, "setup.py")):
            print("Please initialize this project first. Check 'help init'.")
            return
        # check build_report
        self._check_build_report()
        version = manager.get_version(target)
        gurl = create_kurl()
        gurl.token = getpass.getpass("Your Github Token: ")
        print("Processing...")
        print("")
        try:
            data = manager.publish(gurl, target, app_pkg)
        except manager.ReleaseError as e:
            print("Failed to publish the distribution package")
        except manager.AnteReleaseHookError:
            print("Error while running the pre-publishing hook")
        except manager.PostReleaseHookError:
            print("Error while running the post-publishing hook")
        except manager.InvalidReleaseInfoError:
            msg = ("Missing mandatory elements in ",
                   "$APP_DIR/pyrustic_data/publishing.json")
            print("".join(msg))
        except manager.MissingReleaseInfoError:
            msg = ("Missing 'publishing.json' in ",
                   "'pyrustic_data' folder.\nPlease init your app.")
            print("".join(msg))
        else:
            success = self._interpret_release_data(data)
            if success:
                msg = "Successfully published '{}' v{} !"
                print(msg.format(app_pkg, version))

    def _check_build_report(self):
        build_report_json_path = os.path.join(self._target, self._app_pkg,
                                              "pyrustic_data", "build_report.json")
        if not os.path.exists(build_report_json_path):
            BuildHandler(self._target, self._app_pkg)
        jayson = Jayson(build_report_json_path)
        if jayson.data["released"]:
            BuildHandler(self._target, self._app_pkg)

    def _interpret_release_data(self, data):
        meta_code = data["meta_code"]
        status_code = data["status_code"]
        status_text = data["status_text"]
        if meta_code == 0:
            return True
        if meta_code == 1:
            print("Failed to create release.")
            if status_code:
                print("{} {}".format(status_code, status_text))
            else:
                print(status_text)
            return False
        if meta_code == 2:
            print("Failed to upload asset.\n{} {}".format(status_code,
                                                          status_text))
            return False
        print("Unknown error")
        return False



from pyrustic.manager import constant
from jayson import Jayson
from pyrustic.manager.handler.link_handler import LinkHandler
import os.path


class RelinkHandler:
    """
    Description
    -----------
    Link again the previously linked Target or one of
    recent linked Targets.

    Usage
    -----
    - Description: Link again the previously linked Target
    - Command: relink

    - Description: Link again a recently linked Target with
    its index
    - Command: relink <index>

    Example
    -------
    - Description: Link again a previously linked Target
    - Preliminary: Assume you want to link again the Target
    with index #2 (found the index with the command "recent")
    - Command: relink 2
    """
    def __init__(self, target,
                 app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(args)

    @property
    def target(self):
        return self._target

    def _process(self, args):
        jasonix = Jayson(constant.MANAGER_SHARED_DATA_FILE, readonly=True)
        path = jasonix.data["target"]
        if not jasonix.data["recent"]:
            print("- Empty -")
            return
        if len(args) == 1:
            try:
                index = int(args[0])
                path = list(reversed(jasonix.data["recent"]))[index]
            except Exception as e:
                print("Wrong index")
                return
        elif len(args) > 1:
            print("Wrong usage of this command")
            return
        link_handler = LinkHandler(self._target, self._app_pkg, *[path])
        self._target = link_handler.target

    def _check_path(self, path):
        """ Returns True if the path is valid, else False """
        if not os.path.exists(path):
            print("{}".format(path))
            print("This path doesn't exist")
            return False
        return True



from pyrustic import manager


class AddHandler:
    """
    Description
    -----------
    Use this command to add an empty file, a package or a regular
    folder to the Target.

    Usage
    -----
    - Description: Add a package
    - Command: add <my.pack.age>

    - Description: Add a module
    - Command: add <destination> <file.py>

    - Description: Add multiple files
    - Command: add <destination> <file_1.ext> <file_2.ext>

    - Description: Add a file to the Target's APP_DIR
    - Command: add ./ <file.ext>

    Note: The destination is either a relative path to the
    Target's PROJECT_DIR, or a package name.

    Example
    -------
    - Description: Add a module
    - Preliminary: Assume you want to add "my_view.py"
    to the package 'demo.view'
    - Command: add demo.view my_view.py

    - Description: Add a package
    - Preliminary: Assume you want to add 'my.new.package'
    - Command: add my.new.package

    Note: Please use simple or double quotes as delimiters if a
    string contains space.
    """
    def __init__(self, target,
                 app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._args = args
        self._process(target, app_pkg, args)

    def _process(self, target, app_pkg, args):
        if not target:
            self._print_catalog("missing_target")
            return
        if not app_pkg:
            print("Please init the project first. Check 'help init'.")
            return
        if not args:
            self._print_catalog("incomplete")
            return
        # package or folder
        package_or_folder_name = args[0]
        manager.add(target, package_or_folder_name, *args[1:])
        print("Success !")

    def _print_catalog(self, item, **kwargs):
        message = ""
        if item == "missing_target":
            message = "Please link a Target first. Check 'help target'."
        elif item == "incomplete":
            message = "Incomplete command. Check 'help add'."
        elif item == "package_already_exists":
            message = "This package already exists !"
        elif item == "package_not_exists":
            message = "This package doesn't exist yet."
        elif item == "building_package":
            message = "Building package '{}'".format(kwargs["package_name"])
        elif item == "package_built":
            message = "Package successfully built."
        elif item == "failed_to_build_package":
            message = "Failed to finish the package building."
        elif item == "file_created":
            message = "File '{}' successfully created.".format(kwargs["filename"])
        elif item == "file_creation_failed":
            message = "Failed to create file '{}'.".format(kwargs["filename"])
        elif item == "file_exists":
            message = "The file '{}' already exists.".format(kwargs["filename"])
        elif item == "module_exists":
            message = "The module '{}' already exists.".format(kwargs["module"])
        elif item == "module_created":
            message = "Module '{}' successfully created.".format(kwargs["module"])
        elif item == "folder_already_exists":
            message = "This folder already exists !"
        elif item == "folder_not_exists":
            message = "This folder doesn't exist yet."
        elif item == "creating_folder":
            message = "Creating folder '{}'".format(kwargs["folder"])
        elif item == "folder_created":
            message = "Folder successfully created."
        elif item == "failed_to_create_folder":
            message = "Failed to create the folder."
        print(message)



from pyrustic import manager
from pyrustic.manager.core import funcs


class TargetHandler:
    """
    Description
    -----------
    Use this command to check the currently linked Target.

    Usage
    -----
    - Description: Check the currently linked Target
    - Command: target

    Note: This command also shows additional useful
    information about the Target.
    """
    def __init__(self, target,
                 app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(target, app_pkg, args)

    @property
    def target(self):
        return self._target

    def _process(self, target, app_pkg, args):
        if not target:
            print("None")
            return
        # no arg
        if not args:
            self._show_current_version()
        # set a new version
        elif len(args) == 1:
            self._change_current_version(args[0])
        # wrong usage of this command
        else:
            print("Wrong usage of this command")

    def _show_current_version(self):
        data = funcs.check_project_state(self._target)
        print("[{}] {}".format(self._app_pkg,
                               self._target))
        print("Version: {}".format(manager.get_version(self._target)))
        if data == 1:
            print("")
            print("Not yet initialized project (check 'help init')")
        #elif data == 2:
        #    print("Not yet installed project (think about: 'pip install -e .')")

    def _change_current_version(self, new_version):
        pass



import os
import os.path
import sys
import signal
import traceback
from cmd import Cmd
from pyrustic import manager
from pyrustic.manager.core import pymisc
from pyrustic.manager.handler.link_handler import LinkHandler
from pyrustic.manager.handler.unlink_handler import UnlinkHandler
from pyrustic.manager.handler.relink_handler import RelinkHandler
from pyrustic.manager.handler.target_handler import TargetHandler
from pyrustic.manager.handler.recent_handler import RecentHandler
from pyrustic.manager.handler.init_handler import InitHandler
from pyrustic.manager.handler.run_handler import RunHandler
from pyrustic.manager.handler.add_handler import AddHandler
from pyrustic.manager.handler.build_handler import BuildHandler
from pyrustic.manager.handler.publish_handler import PublishHandler
from pyrustic.manager.handler.hub_handler import HubHandler
#from pyrustic.manager.handler.version_handler import VersionHandler
from pyrustic.manager import constant
try:
    import readline
except ImportError:
    readline = None


# decorator for all commands handlers
def guard(func):
    def obj(self, arg):
        cache = None
        try:
            arg = pymisc.parse_cmd(arg) if isinstance(arg, str) else arg
            cache = func(self, arg)
        except Exception as e:
            print("Oops... Exception occurred !\n")
            print("".join(traceback.format_exception(*sys.exc_info())))
        return cache
    return obj


class Cmdline(Cmd):
    intro = ("""Pyrustic Project Manager\n"""
             + """Version: {}\n""".format(manager.dist_version("pyrustic"))
             + """Type "help" or "?" to list commands. Type "exit" to leave.\n""")

    prompt = "(pyrustic) "

    def __init__(self):
        super().__init__()
        self.__history_size = 420
        self.__history_file = None
        self.__target = None
        self.__app_pkg = None
        self.__setup()

    @property
    def target(self):
        if not self.__target:
            return None
        if not os.path.isabs(self.__target):
            self.__target = None
        return self.__target

    @target.setter
    def target(self, val):
        self.__target = val
        self.__app_pkg = None

    @property
    def app_pkg(self):
        if not self.__app_pkg:
            self.__app_pkg = manager.get_app_pkg(self.target)
        return self.__app_pkg

    @app_pkg.setter
    def app_pkg(self, val):
        self.__app_pkg = val

    @property
    def history_size(self):
        return self.__history_size

    @property
    def history_file(self):
        if not self.__history_file:
            path = os.path.join(constant.MANAGER_SHARED_FOLDER,
                                "cmd_history.txt")
            if not os.path.exists(path):
                with open(path, "w") as file:
                    pass
            self.__history_file = path
        return self.__history_file

    # ========== OVERRIDING ==========

    def preloop(self):
        if readline and self.history_file:
            readline.read_history_file(self.history_file)

    def postloop(self):
        if readline:
            readline.set_history_length(self.history_size)
            readline.write_history_file(self.history_file)

    def precmd(self, line):
        if line == "EOF":
            line = ""
        print("")
        return line

    def postcmd(self, stop, line):
        print("")
        return stop

    def emptyline(self):
        pass

    # ========== COMMANDS ==========

    @guard
    def do_link(self, args):
        link_handler = LinkHandler(self.target, self.app_pkg,
                                   *args)
        self.target = link_handler.target

    @guard
    def do_unlink(self, args):
        unlink_handler = UnlinkHandler(self.target,
                                       self.app_pkg,
                                       *args)
        self.target = unlink_handler.target
        self.app_pkg = unlink_handler.app_pkg

    @guard
    def do_relink(self, args):
        relink_handler = RelinkHandler(self.target,
                                       self.app_pkg,
                                       *args)
        self.target = relink_handler.target

    @guard
    def do_target(self, args):
        TargetHandler(self.target,
                      self.app_pkg,
                      *args)

    @guard
    def do_recent(self, args):
        RecentHandler(self.target,
                      self.app_pkg, *args)

    @guard
    def do_init(self, args):
        InitHandler(self.target,
                    self.app_pkg, *args)

    @guard
    def do_run(self, args):
        RunHandler(self.target,
                   self.app_pkg, *args)

    @guard
    def do_add(self, args):
        AddHandler(self.target, self.app_pkg, *args)

    @guard
    def do_build(self, args):
        BuildHandler(self.target, self.app_pkg, *args)

    @guard
    def do_publish(self, args):
        PublishHandler(self.target, self.app_pkg, *args)

    @guard
    def do_hub(self, args):
        HubHandler(self.target, self.app_pkg, *args)

    #@guard
    #def do_version(self, args):
    #   VersionHandler(self.target, self.app_pkg, *args)

    @guard
    def do_exit(self, args):
        print("Exiting...")
        return True

    # ========== COMMANDS DOC ==========

    def help_link(self):
        print(LinkHandler.__doc__)

    def help_unlink(self):
        print(UnlinkHandler.__doc__)

    def help_relink(self):
        print(RelinkHandler.__doc__)

    def help_target(self):
        print(TargetHandler.__doc__)

    def help_recent(self):
        print(RecentHandler.__doc__)

    def help_init(self):
        print(InitHandler.__doc__)

    def help_run(self):
        print(RunHandler.__doc__)

    def help_add(self):
        print(AddHandler.__doc__)

    def help_build(self):
        print(BuildHandler.__doc__)

    def help_publish(self):
        print(PublishHandler.__doc__)

    def help_hub(self):
        print(HubHandler.__doc__)

    #def help_version(self):
    #    print(VersionHandler.__doc__)

    def help_exit(self):
        print("This command closes the program graciously.")

    def __setup(self):
        # ensure install
        manager.install()
        # Interrupt process (typically CTRL+C or 'delete' char or 'break' key)
        signal_num = signal.SIGINT
        handler = lambda signum, frame, self=self: _exit_handler(self)
        signal.signal(signal_num, handler)


def _exit_handler(cmdline):
    cmdline.onecmd("exit")
    cmdline.postloop()
    sys.exit()



"""Pyrustic Project Manager entry point"""
from pyrustic.manager.cmdline import Cmdline
from pyrustic.manager import oneline
import sys
import os


def main():
    """
    Pyrustic Project Manager launcher
    Opens the interactive loop if there are not command line arguments
    else starts pyrustic.manager.oneline.command and uses os.getcwd() as target
    """
    argv = sys.argv
    if len(argv) > 1:
        oneline.command(argv[1:], target=os.getcwd())
    else:
        Cmdline().cmdloop()


if __name__ == "__main__":
    main()



from pyrustic import manager
import os.path


class RecentHandler:
    """
    Description
    -----------
    List of recent Targets.
    The command "relink" links again the Project Manager with
    the Target at index 0 from the "recent" list.

    Usage
    -----
    - Description: List of recent Targets
    - Command: recent
    """

    def __init__(self, target,
                 app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(args)

    def _process(self, args):
        if args:
            print("Wrong usage of this command")
            return
        recent_list = manager.recent()
        len_recent_list = len(recent_list)
        if len_recent_list == 0:
            print("- Empty -")
        for i, path in enumerate(recent_list):
            print("#{}".format(i))
            print("Name: {}".format(os.path.basename(path)))
            print("Path: {}".format(path))
            if i < len_recent_list - 1:
                print("")



import tkinter as tk
from viewable import Viewable
from megawidget.toast import Toast


class HelloView(Viewable):
    """Graphical equivalent of a Hello World"""
    def __init__(self, app):
        super().__init__()
        self._app = app
        self._root = app.root
        self._body = None
        self._btn_max = None

    def _build(self):
        self._body = tk.Frame(self._root)
        self._root.geometry("500x300")
        # Label
        label = tk.Label(self._body, text="Hello Friend !")
        label.pack(expand=1, fill=tk.BOTH)
        # Footer
        footer = tk.Frame(self._body)
        footer.pack(side=tk.BOTTOM, fill=tk.X)
        # Button Leave
        btn_leave = tk.Button(footer, text="Leave",
                              command=self._on_click_btn_leave)
        btn_leave.pack(side=tk.RIGHT, padx=2, pady=2)
        # Button Maximize
        self._btn_max = tk.Button(footer, text="Maximize",
                                  command=self._on_click_btn_max)
        self._btn_max.pack(side=tk.RIGHT, pady=2)

    def _on_map(self):
        pass

    def _on_destroy(self):
        pass

    def _on_click_btn_max(self):
        self._app.maximize()
        self._btn_max.destroy()

    def _on_click_btn_leave(self):
        toast = Toast(self._body, message="Goodbye Friend !")
        toast.wait_window()
        self._app.exit()



"""Pyrustic Framework entry point"""
import platform
import tkinter as tk
from viewable import Viewable, CustomView
import tkutil
from pyrustic.private.enhance_tk import EnhanceTk


class App:
    """
    Pyrustic Framework's entry point.
    This class should be instantiated inside "$APP_DIR/__main__.py".
    """
    def __init__(self):
        """
        """
        self._package = None
        self._is_running = False
        self._restartable = False
        self._root = tk.Tk()
        self._title = None
        self._theme = None
        self._theme_cache = None
        self._view = None
        self._view_cache = None
        self._center_window = False
        self._geometry = "+0+0"
        self._background = "black"
        self._ignore_theme = False
        self._resizable = (True, True)
        self._exit_handler = None
        self._setup()

    # ============================================
    #                PROPERTIES
    # ============================================
    @property
    def root(self):
        """
        Get the main tk root
        """
        return self._root

    @property
    def title(self):
        """
        Get the title of the app
        """
        return self._title

    @title.setter
    def title(self, val):
        """
        Set the title of the app
        """
        self._title = val

    @property
    def theme(self):
        """
        Get the theme object. A theme is a subclass of
        themebase.Theme.
        """
        return self._theme

    @theme.setter
    def theme(self, val):
        """
        Set the theme object. A theme is a subclass of themebase.Theme.

        If you set None, it will invalidate the previous theme.

        Don't forget to call the method "restart()" or "start()" to apply the change !
        Remember that "start()" should be called only once !
        """
        self._theme_cache = val

    @property
    def view(self):
        """
        Get the view object.
        A view should implement "viewable.Viewable"
        """
        return self._view

    @view.setter
    def view(self, val):
        """
        Set a view object. A view should implement "viewable.Viewable".
        If you set None, the previous view will be destroyed.
        The new view will destroy the previous one if there are a previous one.

        Note: "val" can be a tkinter object or a callable (if u plan to REFRESH the app)
        that accepts the app reference as argument and returns a view or a tkinter object
        """
        self._view_cache = val

    @property
    def geometry(self):
        """ Get the geometry string. """
        return self._geometry

    @geometry.setter
    def geometry(self, val):
        """ Set the geometry string. """
        self._geometry = val

    @property
    def background(self):
        """ Get the background color string. By default: "black" """
        return self._background

    @background.setter
    def background(self, val):
        """ Set the background color string """
        self._background = val

    @property
    def ignore_theme(self):
        """ Get the ignore_theme boolean value """
        return self._ignore_theme

    @ignore_theme.setter
    def ignore_theme(self, val):
        """ Set True to ignore the theming """
        self._ignore_theme = val

    @property
    def resizable(self):
        """ Get the resizable boolean"""
        return self._resizable

    @resizable.setter
    def resizable(self, val):
        """ Set a boolean to allow the app to resize or not """
        self._resizable = val

    @property
    def exit_handler(self):
        """ Get the exit handler, a callable called when the app exits """
        return self._exit_handler

    @exit_handler.setter
    def exit_handler(self, val):
        """ Set the exit handler. It will be called when the app exits.
         The exit handler is just a callable """
        self._exit_handler = val

    @property
    def package(self):
        """ Get the package string name """
        return self._package

    @package.setter
    def package(self, val):
        """ Set the package string name """
        self._package = val

    # ============================================
    #               PUBLIC METHODS
    # ============================================
    def start(self):
        """
        Call this method to start the app.
        It should be called once and put on the last line of the file.
        """
        if self._is_running:
            message = "This method shouldn't be called twice. Please use 'restart' instead"
            raise Error(message)
        self._is_running = True
        self._build()
        # main loop
        try:
            self._root.mainloop()
        except KeyboardInterrupt:
            pass

    def restart(self):
        """
        Call this method to restart the app.
        If you have submitted a new view or a new theme,
        the previous view or theme will be removed and the new one installed
        """
        if not self._is_running:
            return False
        if not self._view_cache or not callable(self._view_cache):
            return False
        if self._theme_cache:
            self._apply_theme(self._theme_cache)
            self._theme_cache = None
        if self._view:
            self._view.destroy()
        self._body = None
        return self._install_view(self._view_cache,
                                  is_refresh=True)

    def exit(self):
        """
        Exit, simply ;-)
        """
        self._root.destroy()

    def maximize(self):
        """
        Maximize the window
        """
        system = platform.system()
        if system == "Linux":
            self._root.attributes("-zoomed", True)
        else:  # for "Darwin" (OSX) and "Window"
            self._root.state("zoomed")

    def center(self):
        """
        Center the window
        """
        self._center_window = True

    # ============================================
    #               PRIVATE METHODS
    # ============================================

    def _setup(self):
        pass

    def _build(self):
        # bind self._on_exit
        self._root.protocol("WM_DELETE_WINDOW", self._root.destroy)
        handler = (lambda event,
                          root=self._root:
                   self._on_exit() if event.widget is root else None)
        self._root.bind("<Destroy>", handler)
        EnhanceTk(self._root)
        # set title, apply config, set theme then install view
        self._set_title()
        self._apply_config()
        self._apply_theme(self._theme_cache)
        self._install_view(self._view_cache)

    def _set_title(self):
        if not self._title:
            self._title = "Application"
        text = "{} | built with Pyrustic".format(self._title)
        self._root.title(text)

    def _apply_config(self):
        self._apply_gui_config()

    def _apply_gui_config(self):
        # geometry
        self._root.geometry(self._geometry)
        # background
        self._root.config(background=self._background)
        # resizable width and height
        resizable_width, resizable_height = self._resizable
        self._root.resizable(width=resizable_width, height=resizable_height)

    def _apply_theme(self, theme):
        if self._theme:
            self._root.option_clear()
        self._theme = theme
        if self._ignore_theme:
            return
        if not self._theme:
            return
        self._theme.target(self._root)

    def _install_view(self, view, is_refresh=False):
        self._view = self._get_view(view)
        body = self._view.build()
        if not body:
            return False
        if isinstance(body, tk.Frame):
            self._view.body.pack(in_=self._root,
                                 expand=1, fill=tk.BOTH)
        elif isinstance(body, tk.Toplevel):
            pass
        else:
            self._view.body.pack(in_=self._root)
        # center
        if not is_refresh and self._center_window:
            tkutil.center_window(self._root)
        return True

    def _get_view(self, view):
        if callable(view):
            view = view(self)
        if isinstance(view, Viewable):
            return view
        #if isinstance(view, type) and issubclass(view, Viewable):
        #    return view()
        if view is None:
            view = tk.Frame(self._root,
                            bg="black",
                            width=350,
                            height=200)
        return CustomView(body=view)

    def _on_exit(self):
        if self._exit_handler:
            val = self._exit_handler()


class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.message = args[0] if args else ""
        super().__init__(self.message)

    def __str__(self):
        return self.message



import os
import os.path
import math
import shutil
import subprocess
from jayson import Jayson


def make_archive(name, src, dest, format="zip"):
    """
    - name is the zipfile name minus extension like 'my_archive_file';
    - src is the root dir of the project to zip like '/path/to/project';
    - dest is the dir where to save the final zip.
    Dest should exist like '/path/to/dest'
    format is "zip" or "tar" or "gztar" or "bztar" or “xztar”. Default to "zip"

    Returns the zipfile path and error. Zipfile could be None or str.
    error could be None or an exception instance
    """
    cache = None
    error = None
    try:
        zipfile = os.path.join(dest, name)
        cache = shutil.make_archive(zipfile, format, root_dir=src, base_dir="../..")
    except Exception as e:
        error = e
    return cache, error


def convert_dotted_path(root, dotted_path, prefix="", suffix=""):
    # returns a dotted package name to a regular pathname
    # example: package_name_to_path("/home/proj", "view.lol", prefix="tests.")
    return os.path.join(root, *((prefix + dotted_path).split("."))) + suffix


def tab_to_space(text, tab_size=4):
    TAB = "\t"
    SPACE = " "
    lines = text.split("\n")
    results = []
    for line in lines:
        cache = str()
        for char in line:
            if char == TAB:
                while len(cache) % tab_size != 0:
                    cache += SPACE
            else:
                cache += char
        results.append(cache)
    return "\n".join(results)


def edit_build_version(app_dir):
    about_json_path = os.path.join(app_dir, "pyrustic_data",
                                   "app.json")
    if not os.path.exists(about_json_path):
        return
    jasonix = Jayson(about_json_path)
    current_version = jasonix.data.get("version", "0.0.1")
    message = "The current version is {}".format(current_version)
    print(message)
    message = "Set a new version or ignore: "
    new_version = input(message)
    if not new_version:
        return
    jasonix.data["version"] = new_version
    jasonix.save()


def script(cmd, cwd=None, interactive=True):
    """
    Execute a cmd. Cmd is either a list of args or a string.
    Example of commands:
        - "something 'path/to/dir'"
        - ["something", "path/to/dir"]
    """
    if isinstance(cmd, str):
        cmd = parse_cmd(cmd)
    stdin = None if interactive else subprocess.DEVNULL
    stdout = None if interactive else subprocess.DEVNULL
    stderr = None if interactive else subprocess.DEVNULL
    process = subprocess.Popen(cmd, stdin=stdin, stdout=stdout,
                               stderr=stderr, cwd=cwd)
    if not interactive:
        return 0
    process.communicate()
    return process.returncode


def parse_cmd(cmd):
    """
    Split the cmd string. Delimiters are: space, simple and double quotes
    """
    SINGLE_QUOTE = "'"
    DOUBLE_QUOTE = "\""
    ESPACE = " "
    result = []
    cache = ""
    quote_context = None
    collect_cache = False
    for char in cmd:
        if quote_context is None:  # outside a quote context
            if char in (SINGLE_QUOTE, DOUBLE_QUOTE):
                quote_context = char
                collect_cache = True
            elif char == ESPACE:
                collect_cache = True
            else:
                cache += char
        else:  # inside a quote context
            if char == quote_context:
                quote_context = None
                collect_cache = True
            else:
                cache += char
        # cache collection
        if collect_cache:
            collect_cache = False
            if cache:
                result.append(cache)
            cache = ""
    if cache:
        result.append(cache)
    return result


def convert_size(size):
    """ Size should be in bytes.
    Return a tuple (float_or_int_val, str_unit) """
    if size == 0:
        return (0, "B")
    KILOBYTE = 1024
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size, KILOBYTE)))
    p = math.pow(KILOBYTE, i)
    result = round(size/p, 2)
    return (result, size_name[i])


def truncate_str(data, max_size=15, ellipsis="..."):
    val = ((data[:max_size] + ellipsis)
            if len(data) > max_size else data)
    return val



import os
import os.path
from jayson import Jayson
from pyrustic import manager


class BuildHandler:
    """
    Description
    -----------
    Use this command to build a distribution package
    that could be published later with the 'publish'
    command.
    The distribution package is a Wheel.

    Usage
    -----
    - Description: Build
    - Command: build

    Hooking
    -------
    This command will run hooks if they exist.
    The legal hooks are:
    - pre_building_hook.py
    - post_building_hook.py
    """

    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._version = None
        self._pre_building_hook = None
        self._post_building_hook = None
        self._process(target, app_pkg)

    def _process(self, target, app_pkg):
        if target is None:
            print("Please link a Target first. Check 'help target'.")
            return
        if not os.path.exists(os.path.join(target, "setup.py")):
            print("Please initialize this project first. Check 'help init'.")
            return
        if not self._run_tests():
            return
        version = self._ensure_version()
        try:
            manager.build(target, app_pkg)
        except manager.BuildError:
            print("Failed to build a distribution package")
        except manager.AnteBuildHookError:
            print("Error while running the pre-building hook")
        except manager.PostBuildHookError:
            print("Error while running the post-building hook")
        else:
            print("Successfully built '{}' v{} !".format(app_pkg, version))

    def _run_tests(self):
        if not manager.ask_for_confirmation("Do you want to run tests ?",
                                     "n"):
            print("")
            return True
        print("")
        print("Running tests...")
        test_success, test_result = manager.run_tests(self._target)
        if test_success is None:
            return True
        if test_success:
            print("Testing passed\n")
            return True
        else:
            print("Testing failed")
            print("")
            print(test_result)
            return False

    def _ensure_version(self):
        cur_version = manager.get_version(self._target)
        cache = os.path.join(self._target, self._app_pkg,
                             "pyrustic_data", "build_report.json")
        if not os.path.exists(cache):
            return cur_version
        jasonix = Jayson(cache, readonly=True)
        latest_build_ver = jasonix.data["app_version"]
        version = cur_version
        if latest_build_ver == cur_version:
            print("Current version: {}".format(cur_version))
            print("Type 'maj', 'min' or 'rev' for a fast version increment.")
            print("Ignore it to make an implicit 'rev' increment.")
            new_version = input("Submit a new version: ")
            print("")
            if not new_version:
                new_version = "rev"
            version = manager.interpret_version(cur_version, new_version)
            manager.set_version(self._target, version)
        return version



import os
import os.path
from jayson import Jayson
from pyrustic import manager
from pyrustic.manager.core import github_client


def publish(kurl, target, app_pkg):
    version = manager.get_version(target)
    # get publishing hooks
    ante_release_hooks, post_release_hooks = _get_hooks(target, app_pkg)
    # execute ante_release_hook.py
    for hook in ante_release_hooks:
        if not _run_release_hook(target, app_pkg, version, hook):
            raise manager.AnteReleaseHookError
    # publish
    try:
        data = _publish(kurl, target, app_pkg, version)
    except Exception as e:
        raise manager.ReleaseError
    if data["meta_code"] != 0:
        return data
    # execute post_publishing_hook.py
    for hook in post_release_hooks:
        if not _run_release_hook(target, app_pkg, version, hook):
            raise manager.PostReleaseHookError
    return data


def _get_hooks(target, app_pkg):
    hooking_json = os.path.join(target,
                            app_pkg,
                            "pyrustic_data",
                            "hooking.json")
    if not os.path.exists(hooking_json):
        return [], []
    jayson = Jayson(hooking_json)
    ante_release_hooks = jayson.data.get("ante_release", [])
    post_release_hooks = jayson.data.get("post_release", [])
    return ante_release_hooks, post_release_hooks


def _run_release_hook(target, app_pkg, version, hook):
    if not hook:
        return True
    args = ["-m", hook, target,
            app_pkg, version]
    code = manager.run(target, *args)
    if code == 0:
        return True
    else:
        return False


def _publish(kurl, target, app_pkg, version):
    release_info_json_path = os.path.join(target, app_pkg,
                                        "pyrustic_data",
                                        "release_info.json")
    build_report_json_path = os.path.join(target, app_pkg,
                                          "pyrustic_data",
                                          "build_report.json")
    if not os.path.exists(release_info_json_path):
        raise manager.MissingReleaseInfoError
    jayson = Jayson(release_info_json_path)
    owner = jayson.data.get("owner", None)
    repository = jayson.data.get("repository", None)
    release_name = jayson.data.get("release_name", None)
    tag_name = jayson.data.get("tag_name", None)
    target_commitish = jayson.data.get("target_commitish", None)
    description = jayson.data.get("description", None)
    prerelease = jayson.data.get("is_prerelease", None)
    draft = jayson.data.get("is_draft", None)
    upload_asset = jayson.data.get("upload_asset", None)
    asset_name = jayson.data.get("asset_name", None)
    asset_path = os.path.join(target, "dist", asset_name)
    asset_path = None if not os.path.exists(asset_path) else asset_path
    asset_label = jayson.data.get("asset_label", None)
    if (not release_name or not tag_name
            or not asset_path or not owner or not repository):
        raise manager.InvalidReleaseInfoError
    github_release = github_client.Release(kurl, owner, repository)
    data = github_release.publish(release_name, tag_name,
                                  target_commitish, description,
                                  prerelease, draft, upload_asset,
                                  asset_path, asset_name, asset_label)
    if data["meta_code"] == 201:
        if os.path.exists(build_report_json_path):
            jayson = Jayson(build_report_json_path)
            if version == jayson.data["version"]:
                jayson.data["released"] = True
                jayson.save()
    return data


class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.message, self.code = (args[0], args[1]) if args else ("", None)
        super().__init__(self.message)

    def __str__(self):
        return self.message



import os.path


# == Shared data
# PyrusticData
SHARED_PYRUSTIC_DATA = os.path.join(os.path.expanduser("~"),
                                    "PyrusticData")
# Manager
MANAGER_SHARED_FOLDER = os.path.join(SHARED_PYRUSTIC_DATA, "manager")
MANAGER_SHARED_DATA_FILE = os.path.join(MANAGER_SHARED_FOLDER,
                                        "manager_shared_data.json")
MANAGER_CACHE_FOLDER = os.path.join(MANAGER_SHARED_FOLDER, "cache")














# ============ TODO : delete -->

# sqleditor constants
SQLEDITOR_SHARED_FOLDER = os.path.join(SHARED_PYRUSTIC_DATA, "rustiql")

SQLEDITOR_CACHE_FOLDER = os.path.join(SQLEDITOR_SHARED_FOLDER, "cache")

SQLEDITOR_SHARED_DATA_FILE = os.path.join(SQLEDITOR_SHARED_FOLDER,
                                          "rustiql_shared_data.json")


# runtest constants
RUNTEST_SHARED_FOLDER = os.path.join(SHARED_PYRUSTIC_DATA, "jupitest")

RUNTEST_CACHE_FOLDER = os.path.join(RUNTEST_SHARED_FOLDER, "cache")

RUNTEST_SHARED_DATA_FILE = os.path.join(RUNTEST_SHARED_FOLDER,
                                        "jupitest_shared_data.json")


# hub constants
HUB_SHARED_FOLDER = os.path.join(SHARED_PYRUSTIC_DATA, "hubway")

HUB_CACHE_FOLDER = os.path.join(HUB_SHARED_FOLDER, "cache")

HUB_SHARED_DATA_FILE = os.path.join(HUB_SHARED_FOLDER,
                                      "hubway_shared_data.json")



import unittest
import os.path
import sys


class LiteTestRunner:
    def __init__(self, path, app_dir):
        self._path = path
        self._app_dir = app_dir

    def run(self, failfast=True):
        reloader = _Reloader()
        reloader.save_state()
        cache = self._run(failfast)
        reloader.restore_state()
        return cache

    def _run(self, failfast):
        if not os.path.exists(self._path):
            return False, "This path doesn't exist"
        test_loader = unittest.TestLoader()
        suite = test_loader.discover(self._path, top_level_dir=self._app_dir)
        result = unittest.TestResult()
        try:
            result.startTestRun()
            result.failfast = failfast
            suite.run(result)
        except Exception as e:
            return False, e
        finally:
            result.stopTestRun()
        if result.wasSuccessful():
            return True, None
        else:
            return False, self._stringify_result(result)

    def _stringify_result(self, result):
        data = []
        if result.errors:
            for error in result.errors:
                cache = "{}\n{}".format(error[0], error[1])
                data.append(cache)
        if result.failures:
            for failure in result.failures:
                cache = "{}\n{}".format(failure[0], failure[1])
                data.append(cache)
        if result.unexpectedSuccesses:
            for expected_failure in result.expectedFailures:
                cache = "{}\n{}".format(expected_failure[0],
                                            expected_failure[1])
                data.append(cache)

        return "".join(data)


class _Reloader:
    def __init__(self):
        self._state = None

    def save_state(self):
        self._state = sys.modules.copy()

    def restore_state(self):
        for x in sys.modules.copy().keys():
            if not x in self._state:
                del sys.modules[x]



import os
import os.path
import pkgutil
from jayson import Jayson
from pyrustic.manager.core import funcs


def init(target, app_pkg):
    # create package
    _make_packages(target, app_pkg)
    # create folders
    _make_folders(target, app_pkg)
    # add files
    _add_files(target, app_pkg)
    # add json data files
    _add_json_data_files(target, app_pkg)


def _make_packages(target, app_pkg):
    hooking_pkg = "{}.hooking".format(app_pkg)
    packages = (app_pkg, hooking_pkg, "tests")
    for package in packages:
        funcs.build_package(target, package)


def _make_folders(target, app_pkg):
    # folders to make inside app_pkg
    folders = ("pyrustic_data",)
    for folder in folders:
        path = os.path.join(target, app_pkg, folder)
        if os.path.exists(path):
            continue
        os.mkdir(path)


def _add_files(target, app_pkg):
    resource_prefix = "manager/template/"
    app_dir = os.path.join(target, app_pkg)
    # add VERSION
    resource = resource_prefix + "version_template.txt"
    dest_path = os.path.join(target, "VERSION")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add __main__.py
    resource = resource_prefix + "main_template.txt"
    dest_path = os.path.join(app_dir, "__main__.py")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    data = data.format(title=app_pkg)
    _add_file(dest_path, data)
    # add .gitignore
    resource = resource_prefix + "gitignore_template.txt"
    dest_path = os.path.join(target, ".gitignore")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add LICENSE
    resource = resource_prefix + "license_template.txt"
    dest_path = os.path.join(target, "LICENSE")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add README.md
    resource = resource_prefix + "readme_template.txt"
    dest_path = os.path.join(target, "README.md")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add MANIFEST.in
    resource = resource_prefix + "manifest_template.txt"
    dest_path = os.path.join(target, "MANIFEST.in")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    data = data.format(app_pkg=app_pkg)
    _add_file(dest_path, data)
    # add setup.py
    resource = resource_prefix + "setup_py_template.txt"
    dest_path = os.path.join(target, "setup.py")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add setup.cfg
    resource = resource_prefix + "setup_cfg_template.txt"
    dest_path = os.path.join(target, "setup.cfg")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    data = data.format(project_name=os.path.basename(target),
                       app_pkg=app_pkg)
    _add_file(dest_path, data)
    # add pyproject.toml
    resource = resource_prefix + "pyproject_template.txt"
    dest_path = os.path.join(target, "pyproject.toml")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add LATEST_RELEASE.Md
    resource = resource_prefix + "latest_release_template.txt"
    dest_path = os.path.join(target, "LATEST_RELEASE.md")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add CHANGELOG.Md
    resource = resource_prefix + "changelog_template.txt"
    dest_path = os.path.join(target, "CHANGELOG.md")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add ante_build_hook.py
    resource = resource_prefix + "ante_build_hook_template.txt"
    dest_path = os.path.join(target, app_pkg,
                             "hooking",
                             "ante_build_hook.py")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add post_build_hook.py
    resource = resource_prefix + "post_build_hook_template.txt"
    dest_path = os.path.join(target, app_pkg,
                             "hooking",
                             "post_build_hook.py")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add ante_release_hook.py
    resource = resource_prefix + "ante_release_hook_template.txt"
    dest_path = os.path.join(target, app_pkg,
                             "hooking",
                             "ante_release_hook.py")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)
    # add post_release_hook.py
    resource = resource_prefix + "post_release_hook_template.txt"
    dest_path = os.path.join(target, app_pkg,
                             "hooking",
                             "post_release_hook.py")
    data = pkgutil.get_data("pyrustic", resource).decode("utf-8")
    _add_file(dest_path, data)


def _add_json_data_files(target, app_pkg):
    local_pyrustic_data_folder = os.path.join(target,
                                              app_pkg,
                                              "pyrustic_data")
    resource_prefix = "manager/default_json/pyrustic_data/"
    # add hooking.json
    path = os.path.join(local_pyrustic_data_folder, "hooking.json")
    default_resource = resource_prefix + "hooking_default.json"
    data = pkgutil.get_data("pyrustic", default_resource)
    if not os.path.exists(path):
        with open(path, "wb") as file:
            file.write(data)
    jayson = Jayson(path)
    jayson.data["ante_build"] = ["{}.hooking.ante_build_hook".format(app_pkg)]
    jayson.data["post_build"] = ["{}.hooking.post_build_hook".format(app_pkg)]
    jayson.data["ante_release"] = ["{}.hooking.ante_release_hook".format(app_pkg)]
    jayson.data["post_release"] = ["{}.hooking.post_release_hook".format(app_pkg)]
    jayson.save()
    # add release_info.json
    path = os.path.join(local_pyrustic_data_folder,
                        "release_info.json")
    default_resource = resource_prefix + "release_info_default.json"
    data = pkgutil.get_data("pyrustic", default_resource)
    if not os.path.exists(path):
        with open(path, "wb") as file:
            file.write(data)
    # add hubstore.json
    path = os.path.join(local_pyrustic_data_folder, "hubstore.json")
    default_resource = resource_prefix + "hubstore_default.json"
    data = pkgutil.get_data("pyrustic", default_resource)
    if not os.path.exists(path):
        with open(path, "wb") as file:
            file.write(data)


def _add_file(path, data):
    if os.path.exists(path):
        return
    with open(path, "w") as file:
        file.write(data)
    
        
class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.message = args[0] if args else ""
        super().__init__(self.message)

    def __str__(self):
        return self.message



from pyrustic import manager


class InitHandler:
    """
    Description
    -----------
    Use this command to initialize your project.
    The Project Manager will install a basic
    project structure in the Target.

    Usage
    -----
    - Description: Init your project
    - Command: init
    """
    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(args)

    def _process(self, args):
        if not self._target:
            print("You should link a Target first. Check 'help target'.")
            return
        if args:
            print("Wrong usage of this command. Check 'help init'.")
            return
        # ask for app_pkg
        manager.init(self._target, self._app_pkg)
        print("Successfully initialized !")



import os
import os.path
import shutil
from pyrustic import manager
from pyrustic.manager import constant
from jayson import Jayson
from kurl import Kurl


def check_project_state(target):
    """ Target is the path to the project
    Return:
        0: all right
        1: uninitialized project
        2: uninstalled project
    """
    app_pkg = os.path.basename(target)
    setup_cfg = os.path.join(target, "setup.cfg")
    if not os.path.exists(setup_cfg):
        return 1
    data = manager.dist(app_pkg)
    if not data:
        return 2
    return 0


def wheels_assets(target):
    dist_folder = os.path.join(target,
                               "dist")
    if not os.path.exists(dist_folder):
        return []
    assets = []
    for item in os.listdir(dist_folder):
        _, ext = os.path.splitext(item)
        if ext != ".whl":
            continue
        path = os.path.join(dist_folder, item)
        if not os.path.isfile(path):
            continue
        assets.append(item)
    assets = _sort_wheels_names(assets)
    assets.reverse()
    return assets


def copyto(src, dest):
    """
    Please make sure that DEST doesn't exist yet !
    Copy a file or contents of directory (src) to a destination file or folder (dest)
    """
    if not os.path.exists(src) or os.path.exists(dest):
        return False
    if os.path.isdir(src):
        try:
            shutil.copytree(src, dest)
        except Exception as e:
            return False
    else:
        try:
            shutil.copy2(src, dest)
        except Exception as e:
            return False
    return True


def moveto(src, dest):
    """
    If the DEST exists:
        * Before moveto *
        - /home/lake (SRC)
        - /home/lake/fish.txt
        - /home/ocean (DEST)
        * Moveto *
        moveto("/home/lake", "/home/ocean")
        * After Moveto *
        - /home/ocean
        - /home/ocean/lake
        - /home/ocean/lake/fish.txt
    Else IF the DEST doesn't exist:
        * Before moveto *
        - /home/lake (SRC)
        - /home/lake/fish.txt
        * Moveto *
        moveto("/home/lake", "/home/ocean")
        * After Moveto *
        - /home/ocean
        - /home/ocean/fish.txt


    Move a file or directory (src) to a destination folder (dest)
    """
    if not os.path.exists(src) or os.path.exists(dest):
        return False
    try:
        shutil.move(src, dest)
    except Exception as e:
        return False
    return True


def package_name_to_path(target, package_name, prefix=""):
    # returns a dotted package name to a regular pathname
    # example: package_name_to_path("/home/proj", "view.lol", prefix="tests.")
    return os.path.join(target, *((prefix + package_name).split(".")))


def build_package(target, package_name, prefix=""):
    """
    Literally build a package, returns None or the string pathname
    package represented by prefix must already exist
    """
    splitted = package_name.split(".")
    dir = package_name_to_path(target, prefix) if prefix else target
    for item in splitted:
        dir = os.path.join(dir, item)
        if os.path.exists(dir):
            continue
        try:
            os.mkdir(dir)
            with open(os.path.join(dir, "__init__.py"), "w") as file:
                pass
        except Exception as e:
            return None
    if not os.path.isdir(dir):
        return None
    return dir


def module_name_to_class(module_name):
    """
    Convert a module name like my_module.py to a class name like MyModule
    """
    name = os.path.splitext(module_name)[0]
    # ...
    if not "_" in name:
        return strictly_capitalize(name)
    else:
        splitted = name.split("_")
        cache = []
        for x in splitted:
            cache.append(strictly_capitalize(x))
        return "".join(cache)


def strictly_capitalize(string):
    # I don't remember why I haven't used str.capitalize()
    return string[0].upper() + string[1:]


def get_root_from_package(package_name):
    """
    Return the root from a dotted package name.
    Example the root here "my.package.is.great" is "my".
    """
    splitted = package_name.split(".")
    root = None
    for x in splitted:
        if x == "" or x.isspace():
            continue
        root = x
        break
    return root


def get_manager_jasonix(readonly=True):
    jasonix = Jayson(constant.MANAGER_SHARED_DATA_FILE,
                      readonly=readonly)
    return jasonix


def create_kurl():
    headers = {"Accept": "application/vnd.github.v3+json",
               "User-Agent": "Pyrustic"}
    kurl = Kurl(headers=headers)
    return kurl


def get_hub_url(res):
    target = "https://api.github.com"
    return "{}{}".format(target, res)


def _sort_wheels_names(data):
    cache = list()
    for name in data:
        version = name.split("-")[1]
        cache.append((version, name))
    cache.sort(key=lambda s: [int(i) for i in s[0].split('.')])
    return [name for version, name in cache]



from pyrustic.manager.handler.link_handler import LinkHandler
from pyrustic.manager.handler.unlink_handler import UnlinkHandler
from pyrustic.manager.handler.relink_handler import RelinkHandler
from pyrustic.manager.handler.target_handler import TargetHandler
from pyrustic.manager.handler.recent_handler import RecentHandler
from pyrustic.manager.handler.init_handler import InitHandler
from pyrustic.manager.handler.run_handler import RunHandler
from pyrustic.manager.handler.add_handler import AddHandler
from pyrustic.manager.handler.build_handler import BuildHandler
from pyrustic.manager.handler.publish_handler import PublishHandler
from pyrustic.manager.handler.hub_handler import HubHandler
from pyrustic.manager.handler.version_handler import VersionHandler
from pyrustic.manager import pymisc, get_app_pkg
import os


HANDLERS = {"add": AddHandler, "build": BuildHandler,
            "hub": HubHandler, "init": InitHandler,
            "link": LinkHandler, "publish": PublishHandler,
            "recent": RecentHandler, "relink": RelinkHandler,
            "run": RunHandler, "target": TargetHandler,
            "unlink": UnlinkHandler, "version": VersionHandler}


def command(line=None, target=None):
    """
    Param:
        - line is a string or a list. Example "link /home/project" or ["link", "/home/proj"]
        - target is a path string
    """
    args = None
    if not line:
        return
    if isinstance(line, str):
        args = pymisc.parse_cmd(line)
    else:
        args = line
    app_pkg = get_app_pkg(target)
    if not target:
        target = os.getcwd()
    operation = args[0]
    args = args[1:]
    if operation == "help":
        _show_help(args)
        return
    try:
        HANDLERS[operation](target, app_pkg, *args)
    except KeyError:
        print("Unknown operation")
    except Exception as e:
        raise e
        print("error: ", e)
        print("Oops an error occurred")


def _show_help(args):
    if not args:
        print("Type 'help <operation>' to access the doc.")
        print("Available operations: ")
        for key in HANDLERS.keys():
            print(" - {}".format(key))
        return
    if len(args) > 1:
        print("Incorrect usage of the help command.")
        return
    operation = args[0]
    try:
        print(HANDLERS[operation].__doc__)
    except KeyError:
        print("Unknown operation")
    except Exception:
        print("Oops an error occurred")



import os
import os.path
import tkinter as tk
from tkinter import filedialog
from pyrustic import manager
from pyrustic.manager.core import funcs


class LinkHandler:
    """
    Description
    -----------
    Link your Target project to the Project Manager.

    Usage
    -----
    - Description: Open the directory chooser
    - Command: link

    - Description: Link a Target
    - Command: link </path/to/target/project>

    """
    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(args)

    @property
    def target(self):
        return self._target

    def _process(self, args):
        path = None
        # no args, so open the folder chooser
        if not args:
            path = self._open_folder_chooser()
        # more than 1 arg isn't allowed
        elif len(args) > 1:
            print("Wrong usage of this command")
            return
        # 1 arg submitted: the path
        else:
            path = args[0]
        # invalid path
        if not self._check_path(path):
            return
        # linking
        path = os.path.abspath(path)
        self._link_to(path)

    def _open_folder_chooser(self):
        #initialdir = os.getcwd()
        initialdir = os.path.expanduser("~")
        root = tk.Tk()
        root.withdraw()
        path = filedialog.askdirectory(initialdir=initialdir,
                                            title="Select your project")
        root.destroy()
        if not isinstance(path, str) or not path:
            return
        return path

    def _check_path(self, path):
        """ Returns True if the path is valid, else False """
        if not path:
            print("You haven't submitted a path")
            return False
        if not os.path.exists(path):
            print("This path doesn't exist")
            return False
        return True

    def _link_to(self, path):
        manager.link(path)
        self._target = path
        print("Successfully linked !")
        app_pkg = os.path.basename(path)
        print("[{}] {}".format(app_pkg, path))
        data = funcs.check_project_state(path)
        if data == 1:
            print("")
            print("Not yet initialized project (check 'help init')")



import os
import os.path
import pkgutil
from jayson import Jayson
from pyrustic.manager.constant import SHARED_PYRUSTIC_DATA, MANAGER_SHARED_DATA_FILE


def install():
    # mkdir $HOME/PyrusticData/manager
    _make_shared_dir()
    # add default shared data
    _add_default_shared_data()


def _make_shared_dir():
    path = os.path.join(SHARED_PYRUSTIC_DATA, "manager")
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except Exception as e:
            msg = ("Failed to create the",
                   "shared PyrusticData/manager",
                   "directory inside $HOME")
            raise Error(" ".join(msg))


def _add_default_shared_data():
    resource_prefix = "default_json/PyrusticData"
    resource_json = "manager_shared_data_default.json"
    resource = "/".join((resource_prefix, resource_json))
    data = pkgutil.get_data("pyrustic.manager", resource)
    path = os.path.join(SHARED_PYRUSTIC_DATA, "manager",
                        "manager_shared_data.json")
    try:
        Jayson(MANAGER_SHARED_DATA_FILE, default=data)
    except Exception as e:
        raise Error("Failed to initialize 'manager_shared_data.json'")


class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.message = args[0] if args else ""
        super().__init__(self.message)

    def __str__(self):
        return self.message



from kurl import dict_to_json_body
from pyrustic.manager.core import funcs
import urllib.parse


class Release:
    def __init__(self, gurl, owner, repo):
        self._gurl = gurl
        self._owner = owner
        self._repo = repo

    def publish(self, name, tag_name,
                target_commitish, description,
                prerelease, draft, upload_asset,
                asset_path, asset_name, asset_label):
        """
        Return {"meta_code":, "status_code", "status_text", "data"}
        meta code:
            0- success
            1- failed to create release (check 'status_code', 'status_text')
            2- failed to upload asset (check 'status_code', 'status_text')
        """
        publishing_result = {"meta_code": None, "status_code": None,
                            "status_text": None, "data": None}
        # == create release
        response = self._create_release(self._owner, self._repo, name, tag_name,
                                        target_commitish, description,
                                        prerelease, draft)
        code = response.code
        if code != 201:  # if the release is already created, code: 422 and json is None
            publishing_result["meta_code"] = 1
            publishing_result["status_code"] = code
            publishing_result["status_text"] = response.status[1]
            return publishing_result
        if not upload_asset:
            publishing_result["meta_code"] = 0
            return publishing_result
        # == upload asset
        upload_url = response.json["upload_url"]
        response = self._upload_asset(upload_url, asset_path,
                                      asset_name, asset_label)
        code = response.code
        if code != 201:
            publishing_result["meta_code"] = 2
            publishing_result["status_code"] = code
            publishing_result["status_text"] = response.status[1]
            return publishing_result
        publishing_result["meta_code"] = 0
        return publishing_result

    def _create_release(self, owner, repo, name, tag_name, target_commitish,
                        description, prerelease, draft):
        res = "/repos/{}/{}/releases".format(owner, repo)
        body = {"tag_name": tag_name, "target_commitish": target_commitish,
                "name": name, "body": description, "draft": draft,
                "prerelease": prerelease}
        body = dict_to_json_body(body)
        response = self._gurl.request(funcs.get_hub_url(res),
                                      body=body, method="POST")
        return response

    def _upload_asset(self, upload_url, path, name, label):
        url = upload_url.replace("{?name,label}", "?{}")
        parameters = urllib.parse.urlencode({"label": label, "name": name})
        url = url.format(parameters)
        with open(path, "rb") as file:
            data = file.read()
        header = {"Content-Type": "application/zip"}
        response = self._gurl.request(url, body=data, method="POST",
                                      headers=header)
        return response


def repo_description(gurl, owner, repo):
    """
    Returns: (status_code, status_text, data)
    data = {"created_at": date, "description": str,
            "stargazers_count": int, "subscribers_count": int}
    """
    res = "/repos/{}/{}".format(owner, repo)
    response = gurl.request(funcs.get_hub_url(res))
    code = response.code
    json = response.json
    data = {}
    if code == 304:
        json = response.cached_response.json
    if (code in (200, 304)) and json:
        data["description"] = json["description"]
        date = json["created_at"]
        data["created_at"] = _badass_iso_8601_date_parser(date)
        data["stargazers_count"] = json["stargazers_count"]
        data["subscribers_count"] = json["subscribers_count"]
    return *response.status, data


def latest_release(gurl, owner, repo):
    """
    Returns: (status_code, status_text, data}
    data = {"tag_name": str, "published_at": date,
            "downloads_count": int}
    """
    res = "/repos/{}/{}/releases/latest".format(owner, repo)
    response = gurl.request(funcs.get_hub_url(res))
    code = response.code
    json = response.json
    data = {}
    if code == 304:
        json = response.cached_response.json
    if (code in (200, 304)) and json:
        data["tag_name"] = json["tag_name"]
        date = json["published_at"]
        data["published_at"] = _badass_iso_8601_date_parser(date)
        data["downloads_count"] = _downloads_counter(json)
    return *response.status, data


def latest_releases_downloads(gurl, owner, repo, maxi=10):
    """
    Returns: (status_code, status_text, data}
    data = int, downloads count
    """
    res = "/repos/{}/{}/releases?per_page={}".format(owner, repo, maxi)
    response = gurl.request(funcs.get_hub_url(res))
    code = response.code
    json = response.json
    data = 0
    if code == 304:
        json = response.cached_response.json
    if (code in (200, 304)) and json:
        for release in json:
            data += _downloads_counter(release)
    return *response.status, data


def _downloads_counter(json):
    count = 0
    for asset in json["assets"]:
        count += asset["download_count"]
    return count


def _badass_iso_8601_date_parser(date):
    # YYYY-MM-DDTHH:MM:SSZ
    date = date.rstrip("Z")
    date_part, time_part = date.split("T")
    months = ("Jan", "Feb", "March", "April", "May", "June", "July",
             "Aug", "Sept", "Oct", "Nov", "Dec")
    year, month, day = date_part.split("-")
    text = "{} {} {} at {}".format(day, months[int(month) - 1], year, time_part)
    return text



import tkinter as tk


class EnhanceTk:
    """ Ctrl-a in an Entry or Text doesn't make a text selection by default.
        Now it's possible ! Enjoy ! """

    def __init__(self, root_tk):
        root_tk.bind_class("Entry", "<Control-a>", self._select_all_in_entry, "+")
        root_tk.bind_class("Text", "<Control-a>", self._select_all_in_text, "+")

    def _select_all_in_entry(self, event):
        widget = event.widget
        widget.focus()
        widget.select_range(0, tk.END)
        widget.icursor(tk.END)

    def _select_all_in_text(self, event):
        widget = event.widget
        widget.tag_add(tk.SEL, "1.0", tk.END)


class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.message = args[0] if args else ""
        super().__init__(self.message)

    def __str__(self):
        return self.message





class UnlinkHandler:
    """
    Description
    -----------
    Use this command to unlink the currently linked Target.

    Usage
    -----
    - Description: Unlink the currently linked Target
    - Command: unlink
    """

    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(args)

    @property
    def target(self):
        return self._target

    @property
    def app_pkg(self):
        return self._app_pkg

    def _process(self, args):
        # args are present: invalid command
        if args:
            print("Wrong usage of this command")
            return
        # valid command
        if self._target:
            self._target = None
            self._app_pkg = None
            print("Successfully unlinked !")
        print("Target: {}".format(self._target))



import os
import os.path
from pyrustic import manager
from pyrustic.manager.core.funcs import create_kurl
from jayson import Jayson


class HubHandler:
    """
    Description
    -----------
    Use this command to retrieve useful information
    from a Github repository.

    Usage
    -----
    - Description: Retrieve useful information from
    this repository https://github.com/pyrustic/demo
    - Command: hub pyrustic/demo
    """
    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._args = args
        self._process(args)

    def _process(self, args):
        if not args:
            cache = self._default_owner_repo()
            if cache:
                text = "{}/{}".format(*cache)
                args = [text]
            else:
                print("Wrong usage. Check 'help hub'.")
                return
        elif len(args) > 1:
            print("Wrong usage. Check 'help hub'.")
            return
        owner_repo = args[0].split("/")
        if len(owner_repo) != 2:
            print("Incorrect request. Check 'help hub'.")
            return
        owner, repo = owner_repo
        print("https://github.com/{}/{}\n".format(owner, repo))
        gurl = create_kurl()
        if not self._show_repo_description(gurl, owner, repo):
            return
        if not self._show_latest_release(gurl, owner, repo):
            return
        if not self._show_latest_releases_downloads(gurl, owner, repo):
            return

    def _show_repo_description(self, gurl, owner, repo):
        cache = manager.github_repo_description(gurl, owner, repo)
        status_code, status_text, data = cache
        if status_code not in (200, 304):
            print("Failed to get the repo description")
            if status_code:
                print("{} {}".format(status_code, status_text))
            else:
                print(status_text)
            return False
        self._show_section("Repository description")
        description = data["description"]
        description = "- No description -" if not description else description
        print(description)
        print("Created on {}".format(data["created_at"]))
        stargazers = data["stargazers_count"]
        subscribers = data["subscribers_count"]
        print("{} Stargazer{} and {} Subscriber{}".format(stargazers,
                                                          self._plural(stargazers),
                                                          subscribers,
                                                          self._plural(subscribers)))
        print("")
        return True

    def _show_latest_release(self, gurl, owner, repo):
        cache = manager.github_latest_release(gurl, owner, repo)
        status_code, status_text, data = cache
        if status_code not in (200, 304):
            print("Failed to get the latest release info")
            if status_code:
                print("{} {}".format(status_code, status_text))
            else:
                print(status_text)
            return False
        self._show_section("Latest release")
        print("Tag name: {}".format(data["tag_name"]))
        print("Published on {}".format(data["published_at"]))
        downloads = data["downloads_count"]
        print("{} Download{}".format(downloads,
                                     self._plural(downloads)))
        print("")
        return True

    def _show_latest_releases_downloads(self, gurl, owner, repo):
        cache = manager.github_downloads(gurl, owner, repo)
        status_code, status_text, data = cache
        if status_code not in (200, 304):
            print("Failed to get the latest ten (pre)releases info")
            if status_code:
                print("{} {}".format(status_code, status_text))
            else:
                print(status_text)
            return False
        self._show_section("Latest ten (pre)releases")
        downloads = data
        print("{} Download{}".format(downloads,
                                     self._plural(downloads)))
        print("")
        return True

    def _show_section(self, title):
        count = len(title)
        print(title)
        print("".join(["=" for _ in range(count)]))

    def _plural(self, item):
        item = int(item)
        return "s" if item > 1 else ""

    def _default_owner_repo(self):
        if not self._target and not self._app_pkg:
            return None
        publishing_json_path = os.path.join(self._target, self._app_pkg,
                                            "pyrustic_data",
                                            "publishing.json")
        if not os.path.exists(publishing_json_path):
            return None
        jasonix = Jayson(publishing_json_path)
        if not jasonix.data:
            return None
        owner = jasonix.data["owner"]
        repo = jasonix.data["repo"]
        if not owner or not repo:
            return None
        return owner, repo



import sys
import time
import os
import os.path
import pkgutil
from jayson import Jayson
from pyrustic import manager
from pyrustic.manager.core.funcs import wheels_assets


def build(target, app_pkg):
    # version
    version = manager.get_version(target)
    # build
    ante_build_hooks, post_build_hooks = _get_hooks(target, app_pkg)
    # Run ante build hooks
    for hook in ante_build_hooks:
        if not _run_build_hook(target, app_pkg, version, hook):
            raise manager.AnteBuildHookError
    if sys.executable:
        args = ["setup.py", "--quiet", "sdist", "bdist_wheel"]
        code = manager.run(target, *args)
        if code == 0:
            _gen_build_report(target, app_pkg, version)
            # Run post build hooks
            for hook in post_build_hooks:
                if not _run_build_hook(target, app_pkg, version, hook):
                    raise manager.PostBuildHookError
            return
        raise manager.BuildError


def _run_build_hook(target, app_pkg, version, hook):
    if not hook:
        return True
    args = ["-m", hook, target,
            app_pkg, version]
    code = manager.run(target, *args)
    if code == 0:
        return True
    else:
        return False


def _gen_build_report(target, app_pkg, version):
    pyrustic_data_path = os.path.join(target,
                                      app_pkg,
                                      "pyrustic_data")
    try:
        os.mkdir(pyrustic_data_path)
    except Exception as e:
        pass
    res = "manager/default_json/pyrustic_data/build_report_default.json"
    default_build_report_json = pkgutil.get_data("pyrustic", res)
    build_report_json = os.path.join(pyrustic_data_path,
                                     "build_report.json")
    jasonix = Jayson(build_report_json,
                      default=default_build_report_json)
    jasonix.data["timestamp"] = int(time.time())
    wheels_assets_list = wheels_assets(target)
    wheel_asset = None
    if wheels_assets_list:
        wheel_asset = wheels_assets_list[0]
    jasonix.data["app_version"] = version
    jasonix.data["wheel_asset"] = wheel_asset
    jasonix.data["released"] = False
    jasonix.save()


def _get_hooks(target, app_pkg):
    hooking_json = os.path.join(target,
                                app_pkg,
                                "pyrustic_data",
                                "hooking.json")
    if not os.path.exists(hooking_json):
        return [], []
    jasonix = Jayson(hooking_json)
    ante_build_hooks = jasonix.data.get("ante_build", [])
    post_build_hooks = jasonix.data.get("post_build", [])
    return ante_build_hooks, post_build_hooks


class Error(Exception):
    def __init__(self, *args, **kwargs):
        self.message, self.code = (args[0], args[1]) if args else ("", None)
        super().__init__(self.message)

    def __str__(self):
        return self.message



# ante_release_hook.py generated by Pyrustic Manager
import os
import os.path
import sys
from jayson import Jayson


TEXT = """\
Released with [Pyrustic]({pyrustic_link}).
"""


def get_data():
    """
    Return None or a dict with the keys:
        'script', 'target', 'app_pkg' and 'version'
    """
    items = ("script", "target", "app_pkg", "version")
    data = None
    if len(sys.argv) == len(items):
        data = {item: sys.argv[i] for i, item in enumerate(items)}
    return data


def get_pyrustic_data_path(data):
    target = data["target"]
    app_pkg = data["app_pkg"]
    pyrustic_data_path = os.path.join(target, app_pkg,
                                      "pyrustic_data")
    if not os.path.exists(pyrustic_data_path):
        return None
    return pyrustic_data_path


def get_jayson(pyrustic_data_path, config_name, readonly):
    if not pyrustic_data_path:
        return None
    config_path = os.path.join(pyrustic_data_path,
                               config_name)
    if not os.path.exists(config_path):
        return None
    jayson = Jayson(config_path, readonly=readonly)
    return jayson


def get_release_description(target, owner, repo):
    latest_release_path = os.path.join(target, "LATEST_RELEASE.md")
    cache = None
    try:
        with open(latest_release_path, "r") as file:
            cache = file.read()
    except Exception as e:
        pass
    text = None
    if cache:
        text = cache
    else:
        text = _get_default_release_description(owner, repo)
    return text


def update_release_jayson(target,
                              release_jayson,
                              build_report_jayson,
                              app_pkg, version):
    text = "Missing data in $APP_DIR/pyrustic_data/{}"
    if not build_report_jayson.data:
        print(text.format("build_report.json"))
        return False
    if not release_jayson.data:
        print(text.format("release_info.json"))
        return False
    # update owner
    if not release_jayson.data["owner"]:
        owner = input("Github owner: ")
        print("")
        release_jayson.data["owner"] = owner
    # update repo
    if not release_jayson.data["repository"]:
        repo = os.path.basename(target)
        release_jayson.data["repository"] = repo
    # update name
    repo = release_jayson.data["repository"]
    name = "{} v{}".format(repo, version)
    release_jayson.data["release_name"] = name
    # update tag name
    tag_name = "v{}".format(version)
    release_jayson.data["tag_name"] = tag_name
    # update target commitish
    if not release_jayson.data["target_commitish"]:
        release_jayson.data["target_commitish"] = "master"
    # update description
    owner = release_jayson.data["owner"]
    repository = release_jayson.data["repository"]
    description = get_release_description(target, owner, repository)
    release_jayson.data["description"] = description
    # update is_prerelease
    if not release_jayson.data["is_prerelease"]:
        release_jayson.data["is_prerelease"] = False
    # update is_draft
    if not release_jayson.data["is_draft"]:
        release_jayson.data["is_draft"] = False
    # update upload_asset
    if release_jayson.data["upload_asset"] is None:
        release_jayson.data["upload_asset"] = True
    # update asset name
    asset_name = build_report_jayson.data["wheel_asset"]
    release_jayson.data["asset_name"] = asset_name
    # update asset label
    if not release_jayson.data["asset_label"]:
        release_jayson.data["asset_label"] = "Download the Wheel"
    # save the changes
    release_jayson.save()
    return True


def _get_default_release_description(owner, repo):
    pyrustic_link = "https://github.com/pyrustic/pyrustic"
    text = TEXT.format(pyrustic_link=pyrustic_link)
    return text


def main():
    data = get_data()
    if not data:
        print("Missing sys.argv data")
        sys.exit(1)
    pyrustic_data_path = get_pyrustic_data_path(data)
    # get build_report jayson
    build_report_jayson = get_jayson(pyrustic_data_path,
                                       "build_report.json", True)
    # get release info jayson
    release_jayson = get_jayson(pyrustic_data_path,
                                     "release_info.json", False)
    # update release_info.json
    if not update_release_jayson(data["target"],
                                     release_jayson,
                                     build_report_jayson,
                                     data["app_pkg"],
                                     data["version"]):
        sys.exit(1)
    # exit
    sys.exit(0)


if __name__ == "__main__":
    main()



"""Pyrustic Project Manager API"""

import os
import os.path
import sys
import subprocess
from pyrustic.manager.core import installation, initialization, funcs
from pyrustic.manager.core import building, publishing, github_client
from jayson import Jayson
from pyrustic.manager import constant
from pyrustic.manager.core import funcs
from pyrustic.manager.core.lite_test_runner import LiteTestRunner
if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata
from pyrustic.manager.core import pymisc


def dist_version(package):
    """
    Returns the version of the installed distribution package,
    otherwise returns None.
    """
    return metadata.version(package) if package else None


def install():
    """
    This function does this:
    - create the folder ~/PyrusticData
    - fill the folder with some useful files and sub-folders,
    like a config file for the Project Manager and a "trash"
    folder that apps could use.
    """
    installation.install()


def init(target, app_pkg):
    """
    Initialize a project by creating a basic project structure inside.

    Parameters:
        - target: str, path to the target project
        - app_pkg: str, the application package

    Note: use the function get_app_pkg to extract what could be a
    legal app_pkg from a target.
    """
    initialization.init(target, app_pkg)


def link(target):
    """
    Edits the content of ~/PyrusticData/manager/manager_shared_data.json

    This function will edit the "target" value and the "recent" list.
    """
    jasonix = funcs.get_manager_jasonix(False)
    recent_list = jasonix.data["recent"]
    for i, item in enumerate(recent_list):
        if item == target:
            del recent_list[i]
    recent_list.append(target)
    len_recent_list = len(recent_list)
    max_items = 5
    if len_recent_list > max_items:
        for i in range(len_recent_list - max_items):
            del recent_list[0]
    jasonix.data["target"] = target
    jasonix.save()


def recent():
    """
    Returns a LIFO-sorting list of recently linked targets.
    """
    jasonix = Jayson(constant.MANAGER_SHARED_DATA_FILE)
    recent_list = jasonix.data["recent"]
    return [*reversed(recent_list)]


def add(target, destination, *files):
    """
    Create files and packages in the target project.

    Parameters:
        - target: str, path to the target project
        - destination: str, dotted package name or path relative to the target.
        - files: str, filenames to create.

    Note:
        - The destination can be a dot (in this case, the target is the destination).
        - If a filename already exists, its content is preserved.
        - If destination doesn't exist, it's created.
    """
    # create destination if it doesn't exist
    if destination == "." or ("/" in destination or
                              "\\" in destination):
        destination = os.path.join(target, destination)
        destination = os.path.normpath(destination)
        if not os.path.exists(destination):
            os.makedirs(destination)
    else:
        destination = funcs.build_package(target, destination)
    # add files
    for item in files:
        path = os.path.join(destination, item)
        if os.path.exists(path):
            continue
        with open(path, "w") as file:
            pass


def get_version(target):
    """
    This function read the VERSION file in the target project
    then returns the version (str) of the project.

    Parameters:
         - target: str, path to the target project

    Returns: str, version extracted from $PROJECT_DIR/VERSION or None
    """
    path = os.path.join(target, "VERSION")
    if not os.path.exists(path):
        return None
    with open(path, "r") as file:
        lines = file.readlines()
    if not lines:
        return None
    line = lines[0]
    cache = []
    for char in line:
        if char not in (" ", "\n"):
            cache.append(char)
    return "".join(cache)


def set_version(target, version):
    """
    This function edits the content of $PROJECT_DIR/VERSION

    Parameters:
         - target: str, path to the target project
         - version: str, the version

    Returns:
        - bool, False, if the module version.py is missing
        - bool, True if all right
    """
    if not os.path.exists(target):
        return False
    path = os.path.join(target, "VERSION")
    with open(path, "w") as file:
        file.write(version)
    return True


def interpret_version(cur_version, new_version):
    """
    This function interprets the command to set a new version.

    Parameters:
        - cur_version: str, the current version, the one to alter.
        - new_version: str, the command to set a new version.

    A command can be an actual new version string, or one of the keywords:
     - "maj": to increment the major number of the current version,
     - "min": to increment the minor number of the current version,
     - "rev": to increment the revision number of the current version.

    Returns: The new version as it should be saved in version.py
    """
    if new_version not in ("maj", "min", "rev"):
        return new_version
    cache = cur_version.split(".")
    if not cache:
        return "0.0.1"
    # normalize the size
    if len(cache) == 1:
        cache.extend(["0", "0"])
    elif len(cache) == 2:
        cache.append("0")
    # interpret 'maj', 'min' and 'rev'
    if new_version == "maj":
        number = int(cache[0]) + 1
        cache = [str(number), "0", "0"]
    elif new_version == "min":
        number = int(cache[1]) + 1
        cache = [cache[0], str(number), "0"]
    elif new_version == "rev":
        number = int(cache[2]) + 1
        cache = [cache[0], cache[1], str(number)]
    version = ".".join(cache)
    return version


def build(target, app_pkg):
    """
    Build the target project. A Wheel distribution package is
    created and put in the "dist" folder ($PROJECT_DIR/dist).

    Parameters:
        - target: str, the target
        - app_pkg: str, the application package

    Exceptions:
        - BuildingError: raised if the build fails
        - PreBuildingHookError: raised when pre_building_hook.py exits with non zero
        - PostBuildingHookError: raised when pre_building_hook.py exits with non zero

    Note: this function uses setuptools to build both a source package distribution
    as well as a Wheel package distribution.
    """
    building.build(target, app_pkg)


def run_tests(target):
    """
    Runs the tests in the target.

    Parameters:
        - target: str, path to the target project

    Returns: a tuple (bool, object). The bool indicate the success
    (True) or the failure (False) of the tests.
    The second item in the tuple can be None, an Exception instance, or a string.

    Note: the tests should be located at $PROJECT_DIR/tests
    """
    test_path = os.path.join(target, "tests")
    test_success = None
    test_result = None
    if os.path.exists(test_path):
        test_host = LiteTestRunner(test_path, target)
        test_success, test_result = test_host.run()
    return test_success, test_result


def run(target, *args):
    """
    Run a module or the target project

    Parameters:
        - target: str, path to the target project. Or None. If target is None,
        it will be replaced with the current working directory path.
        - *args: the arguments to run the module with.

    Returns: This function returns the exit code.

    Exception:
        - MissingSysExecutableError: raised when sys.executable is missing.

    Note: this function will block the execution of the thread in
    which it is called till the module exits.
    """
    if not target:
        target = os.getcwd()
    if not sys.executable:
        raise MissingSysExecutableError
    p = subprocess.Popen([sys.executable, *args], cwd=target)
    p.communicate()
    return p.returncode


def ask_for_confirmation(message, default="y"):
    """
    Use this function to request a confirmation from the user.

    Parameters:
        - message: str, the message to display
        - default: str, either "y" or "n" to tell "Yes by default"
        or "No, by default".

    Returns: a boolean, True or False to reply to the request.

    Note: this function will append a " (y/N): " or " (Y/n): " to the message.
    """
    cache = "Y/n" if default == "y" else "y/N"
    user_input = None
    try:
        user_input = input("{} ({}): ".format(message, cache))
    except EOFError as e:
        pass
    if not user_input:
        user_input = default
    if user_input.lower() == "y":
        return True
    return False


def get_app_pkg(target):
    """
    This function extracts the application package name from a target path.
    Basically it extracts the basename from the path then turns dashes "-" into
    "underscores" "_".

    Parameters:
        - target: str, path to the target project

    Returns: str, the application package name.
    """
    if not target:
        return None
    basename = os.path.basename(target)
    cache = basename.split("-")
    app_pkg = "_".join(cache)
    return app_pkg


def github_repo_description(kurl, owner, repo):
    """
    Get useful information about a remote Github repository

    Returns: (status_code, status_text, data)

        The status_code is an int, status_text is a string and
        data = {"created_at": date, "description": str,
                "stargazers_count": int, "subscribers_count": int}
    """
    return github_client.repo_description(kurl, owner, repo)


def github_latest_release(kurl, owner, repo):
    """
    Get from a remote Github repository, useful information about
    the latest release (download count, published at, tag_name)

    Parameters:
        - kurl: python.kurl.Kurl instance
        - owner: str, the Github repository's owner name
        - repo: str, the Github repository's repo name

    Returns: (status_code, status_text, data)

        The status_code is an int, status_text is a string and
        data = {"tag_name": str, "published_at": date,
                "downloads_count": int}
    """
    return github_client.latest_release(kurl, owner, repo)


def github_downloads(kurl, owner, repo, maxi=10):
    """
    Get from a remote Github repository the total downloads count
    of the latest x (maxi) releases assets.

    Parameters:
        - kurl: python.kurl.Kurl instance
        - owner: str, the Github repository's owner name
        - repo: str, the Github repository's repo name
        - maxi: int, number of releases to take in consideration
         (from the latest to the oldest)

    Returns: (status_code, status_text, data)

        The status_code is an int, status_text is a string and
        data is an integer to indicate the downloads count
    """
    return github_client.latest_releases_downloads(kurl, owner, repo, maxi)


def publish(kurl, target, app_pkg):
    """
    Publish the latest built distribution package of the target project.

    Parameters:
        - kurl: a pyrustic.kurl.Kurl instance
        - target: str, path to the target project
        - app_pkg: str, application package

    Exceptions:
        - PublishingError: raised if the publishing failed
        - PrePublishingHookError: raised if pre_publishing_hook.py exits with non zero
        - PrePublishingHookError: raised if pre_publishing_hook.py exits with non zero

    Note: this function uses Github API to create a new release
    in the Github repository, then upload the distribution package
    as an asset.

    The JSON file "publishing.json" located at $APP_DIR/pyrustic_data
    is read to extract useful data to perform this operation.
    """
    return publishing.publish(kurl, target, app_pkg)


def dist(name):
    """
    Use this function to get some info about a distribution package

    Parameters:
        name: the distribution name, example: "wheel"

    Returns: A dict with these keys:
        name, author, author_email, description, home_page,
        maintainer, maintainer_email, version.

    Note: All values in the returned dict are strings.
    """
    metadata_cache = None
    try:
        metadata_cache = metadata.metadata(name)
    except Exception:
        pass
    keys = (("author", "Author"),
            ("author_email", "Author-email"),
            ("description", "Summary"),
            ("home_page", "Home-page"),
            ("maintainer", "Maintainer"),
            ("maintainer_email", "Maintainer-email"),
            ("version", "Version"))
    data = None
    if metadata_cache:
        data = {"name": name}
        for item in keys:
            if item[1] in metadata_cache:
                data[item[0]] = metadata_cache[item[1]]
    return data


class Error(Exception):
    """ """
    def __init__(self, *args, **kwargs):
        """ """
        self.code = 0
        self.message = args[0] if args else ""
        super().__init__(self.message)

    def __str__(self):
        return self.message


class AnteReleaseHookError(Error):
    """ """
    pass


class PostReleaseHookError(Error):
    """ """
    pass


class ReleaseError(Error):
    """ """
    pass


class MissingReleaseInfoError(Error):
    """ """
    pass


class InvalidReleaseInfoError(Error):
    """ """
    pass


class AnteBuildHookError(Error):
    """ """
    pass


class PostBuildHookError(Error):
    """ """
    pass


class BuildError(Error):
    """ """
    pass


class MissingSysExecutableError(Error):
    """ """
    pass



import os
import os.path
from pyrustic import manager


class RunHandler:
    """
    Description
    -----------
    Use this command to run a module.
    The module can be located either in the Target or in
    a regular place where Python stores packages.
    Only dotted name of a module is allowed, so please ignore
    the extension ".py".

    Usage
    -----
    - Description: Run a module
    - Command: run <the.module.name>

    - Description: Run the Target
    - Command: run
    Note: Project Manager will implicitly run APP_DIR/__main__.py

    - Description: Run a module with some arguments
    - Command: run <the.module.name> <argument_1> <argument_2>

    Example
    -------
    - Description: Run the module
    - Preliminary: Assume that "my_view.py" is in the "view" package
    - Command: run view.my_view

    - Description: Run the module with arguments
    - Preliminary: Assume that "my_view.py" is in the "view" package
    - Command: run view.my_view argument_1 "argument 2"

    - Description: Display the Zen of Python
    - Command: run this

    Note: Please use simple or double quotes as delimiters if a string
    contains space.
    """
    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._process(target, app_pkg, args)

    def _process(self, target, app_pkg, args):
        name = None
        if not args and not target:
            self._print_catalog("missing_target")
            return
        if not args and not os.path.exists(os.path.join(target, app_pkg)):
            print("Please initialize this project first. Check 'help init'.")
            return
        if not args and target:
            args = ["-m", app_pkg]
        else:
            args = ["-m", *args]
        name = " ".join(args)
        self._print_catalog("running", module=name)
        print("")
        manager.run(target, *args)

    def _print_catalog(self, item, **kwargs):
        message = ""
        if item == "missing_target":
            message = "Please link a Target first. Check 'help target'."
        elif item == "running":
            message = "Running '{}' ...".format(kwargs["module"])
        elif item == "missing_app_pkg":
            message = "Please init the project first. Check 'help init'."
        print(message)



from pyrustic import manager


class VersionHandler:
    """
    Description
    -----------
    Use this command to check or edit the version of the
    Target project.

    Usage
    -----
    - Description: Check the current version
    - Command: version

    - Description: Set a new version
    - Command: version <sequence>

    - Description: Increment the major number
    - Command: version maj

    - Description: Increment the minor number
    - Command: version min

    - Description: Increment the revision number
    - Command: version rev

    Example
    -------
    - Description: Increment the major number
    - Preliminary: Assume the current version 1.2.3
    - Command: version maj
    - Result: The new version is: 2.0.0

    - Description: Set a version
    - Command: version 2.0.1
    """
    def __init__(self, target, app_pkg, *args):
        self._target = target
        self._app_pkg = app_pkg
        self._version = None
        self._process(args)

    def _process(self, args):
        if not self._target:
            print("You should link a Target first. Check 'help target'.")
            return
        self._version = manager.get_version(self._target)
        # no arg
        if not args:
            self._show_current_version()
        # set a new version
        elif len(args) == 1:
            self._change_current_version(args[0])
        # wrong usage of this command
        else:
            print("Wrong usage of this command")

    def _show_current_version(self):
        print("Project version: {}".format(self._version))

    def _change_current_version(self, new_version):
        new_version = manager.interpret_version(self._version, new_version)
        manager.set_version(self._target, new_version)
        print("Previous value : {}".format(self._version))
        print("Project version: {}".format(new_version))
        self._version = new_version



# post_release_hook.py generated by Pyrustic Manager
import sys
import os
import os.path
from datetime import datetime
import time


def get_data():
    """
    Return None or a dict with the keys:
        'script', 'target', 'app_pkg' and 'version'
    """
    items = ("script", "target", "app_pkg", "version")
    data = None
    if len(sys.argv) == len(items):
        data = {item: sys.argv[i] for i, item in enumerate(items)}
    return data


def get_date():
    """ Returns the current date. Format: Month day, year.
    Example: January 15, 2020
    """
    MONTHS = ("January", "February", "March", "April", "May",
              "June", "July", "August", "September", "October",
              "November", "December")
    dt = datetime.fromtimestamp(time.time())
    text = "{month} {day}, {year}".format(month=MONTHS[dt.month - 1],
                                          day=dt.day, year=dt.year)
    return text


def update_changelog(path, data, version):
    """ Update the file CHANGELOG.md located at path, with data and version """
    if not data:
        return
    cache = "## Version {} of {}\n"
    data.insert(0, cache.format(version, get_date()))
    data.append("\n\n\n")
    data = "".join(data)
    try:
        with open(path, "r+") as file:
            cache = file.readlines()
            cache.insert(0, data)
            file.seek(0)
            file.write("".join(cache))
            file.truncate()
    except Exception as e:
        pass


def main():
    data = get_data()
    if not data:
        print("Missing sys.argv data")
        sys.exit(1)
    # cut LATEST_RELEASE.md content and then log it in CHANGELOG.md
    target = data["target"]
    latest_release_path = os.path.join(target, "LATEST_RELEASE.md")
    try:
        with open(latest_release_path, "r+") as file:
            cache = file.readlines()
            file.seek(0)
            file.write("")
            file.truncate()
    except Exception as e:
        pass
    else:
        changelog_path = os.path.join(target, "CHANGELOG.md")
        update_changelog(changelog_path, cache, data["version"])
    # exit
    sys.exit(0)


if __name__ == "__main__":
    main()



# ante_build_hook.py generated by Pyrustic Manager
import sys


def get_data():
    """
    Return None or a dict with the keys:
        'script', 'target', 'app_pkg' and 'version'
    """
    items = ("script", "target", "app_pkg", "version")
    data = None
    if len(sys.argv) == len(items):
        data = {item: sys.argv[i] for i, item in enumerate(items)}
    return data


def main():
    sys.exit(0)


if __name__ == "__main__":
    main()
